from ...db.schema import Access as DBAccess
from ...db.schema import Role as DBRole
from ...db.schema import Schema
from ..defaults import default_blueprint
from ..field import MutableField, ReverseField
from .build import TableBuilder


def create_schema():
    table_builder = TableBuilder()
    tables = [
        table_builder.build(model)
        for model in default_blueprint.iter_models()
    ]
    roles = [
        role_to_dbrole(role)
        for role in default_blueprint.iter_roles()
    ]
    access = [
        access_to_dbaccess(access)
        for access in default_blueprint.iter_access()
    ]
    schema = Schema(tables, roles=roles, access=access)
    schema.bind()
    return schema


def role_to_dbrole(role):
    dbrole = getattr(role.Meta, 'dbrole', None)
    if not dbrole:
        dbrole = DBRole(
            role.Meta.role,
            parents=[
                role_to_dbrole(parent)
                for parent in role.parents
            ],
            app=role.Meta.app
        )
        role.Meta.dbrole = dbrole
    return dbrole


def access_to_dbaccess(access):
    dbaccess = getattr(access, 'dbaccess', None)
    if not dbaccess:
        dbaccess = DBAccess(
            access.entity.Meta.table,  # TODO: Generalise.
            all=[
                role_to_dbrole(role)
                for role in access.all
            ],
            select=[
                role_to_dbrole(role)
                for role in access.select
            ],
            insert=[
                role_to_dbrole(role)
                for role in access.insert
            ],
            update=[
                role_to_dbrole(role)
                for role in access.update
            ],
            delete=[
                role_to_dbrole(role)
                for role in access.delete
            ],
            app=access.app
        )
        access.dbaccess = dbaccess
    return dbaccess


def model_to_values(model, exclude_fields=None):
    exclude_fields = exclude_fields or ()
    values = {}
    for field_name, field in model.Meta.fields.items():
        if field_name in exclude_fields:
            continue
        if not isinstance(field, (MutableField, ReverseField)):
            continue
        if not hasattr(model, field_name):
            continue
        values[field_name] = field.to_outgoing(
            model,
            getattr(model, field_name)
        )
    return values


def set_values_on_model(values, model):
    for field_name, field in model.Meta.fields.items():
        if not isinstance(field, MutableField):
            continue
        if field_name not in values:
            continue
        setattr(model, field_name, field.from_incoming(
            model,
            values[field_name]
        ))



from psycopg2.sql import SQL

from .expression import Expression


class Multi(Expression):
    def __init__(self, *expressions):
        self.expressions = expressions

    def to_sql(self):
        all_sql = []
        all_args = ()
        for expr in self.expressions:
            sql, args = expr.to_sql()
            all_sql.append(sql)
            all_args += args
        return SQL('; ').join(all_sql), all_args



from ...db.schema import Table
from ..field import MutableField
from .convert import convert_field


class TableBuilder:
    def __init__(self):
        self.processing_models = set()

    def build(self, model):
        if model in self.processing_models:
            # TODO: This is to catch self-referential
            # models. Returning the name means the DB schema can bind
            # it later. Refactor for clarity.
            table_name = f'{model.Meta.table_name}'
            if model.Meta.app:
                # TODO: Also pretty horrible.
                table_name = f'{model.Meta.app.name.lower()}_{table_name}'
            return table_name
        table = getattr(model.Meta, 'table', None)
        if not table:
            self.processing_models.add(model)
            table = self.build_table(model)
            self.processing_models.remove(model)
            model.Meta.table = table
        return table

    def build_table(self, model):
        return Table(
            model.Meta.table_name,
            columns=self.build_all_columns(model),
            app=model.Meta.app,
            uniques=model.Meta.uniques,
            checks=model.Meta.checks,
            indexes=model.Meta.indexes,
            policies=model.Meta.policies
        )

    def build_all_columns(self, model):
        columns = []
        for field in model.Meta.fields.values():
            if not isinstance(field, MutableField):
                continue
            columns.append(self.build_column(field))
        return columns

    def build_column(self, field):
        return convert_field(field, self)



import click

from .main import main

__all__ = ('show',)


@main.group()
@click.pass_context
def app(ctx):
    pass


@app.command()
def show():
    from ..project.project import load_project
    project = load_project()
    project.prepare()
    print(project.apps)



from coolname import generate_slug
from psycopg2.sql import SQL, Identifier
from termcolor import colored

from ...db.connection import cursor
from ...utils import random_ident
from ...utils.feedback import feedback
from .constants import DEPLOYMENT_PREFIX
from .exceptions import KnownError
from .operations import (delete_parameter, get_parameter,
                         get_parameters_by_path, set_parameter)
from .secret import create_secret
from .utils import aws_client


def get_db_name(project, deployment):
    return f'{project.lower()}_{deployment.lower()}'


def get_rds_instance(instance_id, rds=None):
    rds = aws_client('rds')
    response = rds.describe_db_instances(
        Filters=[{
            'Name': 'db-instance-id',
            'Values': [instance_id]
        }]
    )
    return response['DBInstances'][0]


def get_db_url(username, password, address, db_name):
    return f'postgres://{username}:{password}@{address}/{db_name}'


def get_db_url_components(url):
    # TODO: Fix
    return {
        'address': url[url.find('@') + 1:url.rfind('/')],
        'db_name': url[url.rfind('/'):]
    }


def replace_db_name(url, db_name):
    return url[:url.rfind('/') + 1] + db_name


@feedback
def create_instance(url=None, instance_class='db.t2.micro',
                    storage=10, master_username='master', backup_days=0,
                    feedback=None):
    rds = aws_client('rds')
    instance_name = generate_slug(2)
    password = random_ident(36)
    action = 'Use' if url else 'Provision'
    with feedback(f'{action} database instance {colored(instance_name, "blue")}'):
        if not url:
            rds.create_db_instance(
                DBName='master',
                DBInstanceIdentifier=instance_name,
                DBInstanceClass='db.t2.micro',
                Engine='postgres',
                AllocatedStorage=storage,
                StorageType='gp2',
                MasterUsername=master_username,
                MasterUserPassword=password,
                BackupRetentionPeriod=backup_days,
                PubliclyAccessible=True,
                Tags=[{
                    'Key': 'Builder',
                    'Value': 'Polecat'
                }]
            )
            waiter = rds.get_waiter('db_instance_available')
            waiter.wait(DBInstanceIdentifier=instance_name)
            address = get_rds_instance(instance_name)['Endpoint']['Address']
            url = get_db_url('master', password, address, 'master')
        set_parameter(
            f'/polecat/aws/db/instances/{instance_name}/url',
            url
        )
    return instance_name, url


@feedback
def delete_instance(name, feedback=None):
    rds = aws_client('rds')
    with feedback(f'Delete database instance {colored(name, "blue")}'):
        databases = get_parameters_by_path(
            f'/polecat/aws/db/instances/{name}/databases/'
        )
        if databases:
            raise KnownError('instance not empty')
        rds.delete_db_instance(
            DBInstanceIdentifier=name,
            DeleteAutomatedBackups=True,
            SkipFinalSnapshot=True
        )
        waiter = rds.get_waiter('db_instance_deleted')
        waiter.wait(DBInstanceIdentifier=name)
        delete_parameter(
            f'/polecat/aws/db/instances/{name}/url'
        )


def list_instances(feedback=None):
    with feedback('Fetch instances') as fb:
        instances = get_parameters_by_path(f'/polecat/aws/db/instances/')
    return instances.keys()


@feedback
def create_db(project, deployment, instance_name=None, instance_class='db.t2.micro',
              storage=10, master_username='master', backup_days=0,
              feedback=None):
    if not instance_name:
        instance_name, instance_url = create_instance(
            instance_class=instance_class,
            storage=storage,
            master_username=master_username,
            backup_days=backup_days,
            feedback=feedback
        )
    else:
        instance_url = get_parameter(f'/polecat/aws/db/instances/{instance_name}/url')
    address = get_db_url_components(instance_url)['address']
    with feedback(f'Create database for {colored(project, "blue")}/{colored(deployment, "green")}'):
        db_name = get_db_name(project, deployment)
        password = random_ident(36)
        url = get_db_url(db_name, password, address, db_name)
        with cursor(instance_url) as curs:
            curs.execute(
                SQL('CREATE USER {} WITH CREATEROLE PASSWORD \'%s\'' % password).format(
                    Identifier(db_name)
                )
            )
            curs.execute(
                SQL('GRANT {} TO {}').format(
                    Identifier(db_name),
                    Identifier('master')
                )
            )
            curs.execute(
                SQL('CREATE DATABASE {db_name} WITH OWNER {db_name};').format(
                    db_name=Identifier(db_name)
                )
            )
            curs.execute(
                SQL('GRANT ALL PRIVILEGES ON DATABASE {db_name} TO {db_name};').format(
                    db_name=Identifier(db_name)
                )
            )
        with cursor(replace_db_name(instance_url, db_name)) as curs:
            # TODO: Allow specification of other extensions.
            curs.execute((
                'CREATE EXTENSION IF NOT EXISTS chkpass;'
                'CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'
            ))
        set_parameter(
            f'/polecat/aws/db/instances/{instance_name}/databases/{db_name}/url',
            url
        )
        set_parameter(
            DEPLOYMENT_PREFIX.format(project, deployment) + f'databases/{db_name}/url',
            url
        )
        set_parameter(
            DEPLOYMENT_PREFIX.format(project, deployment) + f'databases/{db_name}/instance_name',
            instance_name
        )
    create_secret(
        project,
        deployment,
        'DATABASE_URL',
        url,
        feedback=feedback
    )


@feedback
def list_dbs(project, deployment, feedback=None):
    with feedback(f'List databases for {colored(project, "blue")}/{colored(deployment, "green")}'):
        dbs = get_parameters_by_path(
            DEPLOYMENT_PREFIX.format(project, deployment) + 'databases/'
        )
    return dbs.keys()


@feedback
def delete_db(project, deployment, feedback=None):
    db_name = get_db_name(project, deployment)
    with feedback(f'Delete database for {colored(project, "blue")}/{colored(deployment, "green")}'):
        info = get_parameters_by_path(
            DEPLOYMENT_PREFIX.format(project, deployment) + f'databases/{db_name}/',
        )
        instance_name = info['instance_name']
        instance_url = get_parameter(
            f'/polecat/aws/db/instances/{info["instance_name"]}/url'
        )
        with cursor(instance_url) as curs:
            curs.execute('''
              SELECT pg_terminate_backend(pg_stat_activity.pid)
              FROM pg_stat_activity
              WHERE pg_stat_activity.datname = '{}'
                AND pid <> pg_backend_pid();
            '''.format(db_name))
            curs.execute(SQL('DROP DATABASE {}').format(Identifier(db_name)))
            curs.execute(SQL('DROP USER {}').format(Identifier(db_name)))
        delete_parameter([
            f'/polecat/aws/db/instances/{instance_name}/databases/{db_name}/url',
            DEPLOYMENT_PREFIX.format(project, deployment) + f'databases/{db_name}/url',
            DEPLOYMENT_PREFIX.format(project, deployment) + f'databases/{db_name}/instance_name'
        ])



__all__ = ('SessionVariable',)


class SessionVariable:
    def __init__(self, name, type=None):
        self.name = name
        self.type = type

    def __eq__(self, other):
        return other and self.name == other.name

    def __ne__(self, other):
        return not self.__eq__(other)



import time
from functools import wraps

from .decorator import decorator_with_args


@decorator_with_args
def retry(func, max_retries=2, delay_ms=1000, swallow_error=False):
    @wraps(func)
    def inner(*args, **kwargs):
        cur_retry = 0
        while True:
            try:
                return func(*args, **kwargs)
            except Exception:
                if cur_retry == max_retries:
                    if swallow_error:
                        return
                    raise
                time.sleep(delay_ms/1000)
                cur_retry += 1
    return inner



from jwt import encode, decode
from polecat.core.config import default_config


def jwt(claims):
    return encode(
        claims or {},
        default_config.jwt_secret,
        algorithm='HS256'
    ).decode()


def jwt_decode(token):
    if isinstance(token, str):
        token = token.encode()
    return decode(
        token,
        default_config.jwt_secret,
        algorithms=('HS256',)
    )



from psycopg2.sql import SQL, Identifier

from .expression import Expression


class As(Expression):
    @classmethod
    def factory(cls, expression, alias):
        if not isinstance(expression, As):
            return cls(expression, alias)
        else:
            return expression

    def __init__(self, expression, alias):
        self.expression = expression
        self.alias = alias

    @property
    def root_relation(self):
        return self.expression.root_relation

    def to_sql(self):
        expr_sql, expr_args = self.expression.to_sql()
        as_sql, as_args = self.get_as_sql(expr_sql)
        return (
            as_sql,
            expr_args + as_args
        )

    def get_as_sql(self, expression_sql):
        return SQL('{} AS {}').format(
            expression_sql,
            Identifier(self.alias)
        ), ()

    def get_subrelation(self, name):
        # TODO: This feels like bad design. Not sure how to correct
        # just at the moment, and at least it's clear what it's doing.
        return self.expression.get_subrelation(name)

    def get_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        return self.expression.get_column(name)

    def has_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        return self.expression.has_column(name)

    def push_selection(self, selection=None):
        self.expression.push_selection(selection)



import inspect
from pathlib import Path

from ...utils import to_class


def project_migrations_path():
    from ...project.project import get_active_project
    active_project = get_active_project()
    if active_project is None:
        return Path.cwd()
    return Path(inspect.getfile(to_class(active_project))).parent / 'migrations'



import click

from .config import update_config
from .feedback import HaloFeedback
from .main import main

__all__ = ('initialise',)


@main.group()
@click.pass_context
def aws(ctx):
    pass


@aws.command()
@click.argument('bucket')
def initialise(bucket):
    from ..deploy.aws.initialise import initialise as aws_initialise
    from botocore.exceptions import ClientError
    try:
        aws_initialise(bucket, feedback=HaloFeedback())
    except ClientError as e:
        if e.response['Error']['Code'] == 'IllegalLocationConstraintException':
            print('\nPlease confirm you have correctly setup AWS for use on your machine.')
            print('You may need to specify your AWS credentials in your environment.\n')
        else:
            raise
    update_config({
        'bucket': bucket
    })



from setuptools import find_packages, setup

version = '0.0.10'

setup(
    name='polecat',
    version=version,
    author='Luke Hodkinson',
    author_email='furious.luke@gmail.com',
    maintainer='Luke Hodkinson',
    maintainer_email='furious.luke@gmail.com',
    description='',
    long_description='',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License'
    ],
    packages=find_packages(exclude=['tests', 'extern']),
    include_package_data=True,
    package_data={'': ['*.txt', '*.js', '*.html', '*.*']},
    install_requires=[
        'ujson~=1.35',
        'psycopg2-binary~=2.8.4',
        'graphql-core-next~=1.0.2',
        # TODO: This should be in "cli".
        'polecat-feedback',
        # 'graphql-server-core @ https://github.com/norman-thomas/graphql-server-core/tarball/master',
        'pyjwt~=1.7.1',
        'pytimeparse~=1.1.8',
        'pytz~=2019.1',
        # Used to be in extras...
        'click~=7.0',
        'halo~=0.0.29',
        'termcolor~=1.1.0',
        'cached-property~=1.5.1'
    ],
    extras_require={
        'cli': [
            'pytest~=4.3.1',
            'pytest-asyncio~=0.10.0',
            'pytest-cov~=2.8.1',
            'pytest-polecat',
            'factory-boy~=2.11.1',
            'sanic~=19.3.1',
            'sanic-cors~=0.9.9.post3',
            'boto3~=1.9.125',
            'coolname~=1.1.0',
            'python-Levenshtein~=0.12.0'
        ]
    },
    entry_points={
        'console_scripts': [
            'polecat=polecat.cli.entrypoint:entrypoint'
        ]
    },
    zip_safe=True
)



from polecat.core.context import active_context
from polecat.utils.container import Option, OptionDict, passthrough

active_context().Meta.add_options(
    Option('db', default=passthrough(OptionDict)(
        (
            Option('schema'),
        )
    ))
)



from polecat.core.context import active_context
from polecat.db.decorators import dbcursor
from polecat.db.migration import diff_schemas
from polecat.db.migration import make_migrations as db_make_migrations
from polecat.db.migration import migrate

from .helpers import create_schema


@dbcursor
def sync(migration_paths=None, cursor=None):
    migrate(migration_paths, cursor=cursor)
    # TODO: Build schema from migrations. Should return from above?
    to_schema = create_schema()
    migrations = diff_schemas(to_schema)  # TODO: Add in "from_schema"
    for mgr in migrations:
        mgr.forward()


@active_context
def make_migrations(context=None, **kwargs):
    return db_make_migrations(
        apps=context.registries.app_registry,
        **kwargs
    )



from .passthrough import passthrough  # noqa


class Undefined:
    pass


class Option:
    def __init__(self, name, type=Undefined, default=Undefined):
        self.name = name
        self.default = default
        self.type = type


class OptionDict:
    """ A dictionary-like class in which all possible options are
    specified in advance. Allows for acknowledgement of incorrect
    accesses and suggestions for the correct key.
    """
    def __init__(self, options=None):
        self.options = set()
        self.items = {}
        self.types = {}
        self.add_options(*(options or ()))

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def get(self, key):
        try:
            return self.items[key]
        except KeyError:
            pass
        if key not in self.options:
            raise self.key_error(key)
        raise ValueError(f'No value set for key {key}')

    def set(self, key, value):
        if key not in self.options:
            raise self.key_error(key)
        self.items[key] = self.coerce(key, value)

    def coerce(self, key, value):
        if key in self.types:
            type = self.types[key]
            if type == bool:
                if isinstance(value, str):
                    return not (
                        not len(value) or value.lower()[0] in ('f', 'n', '0')
                    )
            value = type(value)
        return value

    def add_options(self, *options):
        keys_to_add = set()
        defaults_to_add = {}
        types_to_add = {}
        for opt in options:
            keys_to_add.add(opt.name)
            if opt.default != Undefined:
                defaults_to_add[opt.name] = opt.default
            if opt.type != Undefined:
                types_to_add[opt.name] = opt.type
        self.init_defaults(keys_to_add, defaults_to_add)
        self.options.update(keys_to_add)
        self.items.update(defaults_to_add)
        self.types.update(types_to_add)
        return self

    def init_defaults(self, keys_to_add, defaults_to_add):
        pass

    def remove_options(self, *options):
        for opt in options:
            try:
                self.options.remove(opt)
            except KeyError:
                pass
            try:
                del self.items[opt]
            except KeyError:
                pass
        return self

    def key_error(self, key):
        try:
            from Levenshtein import distance
        except ImportError:
            return KeyError(f'{key} not found')
        closest = [
            (distance(key, opt), opt)
            for opt in self.options
        ]
        if closest:
            closest = min(closest, key=lambda x: x[0])[1]
            return KeyError(f'{key} not found, did you mean {closest}?')
        else:
            return KeyError(f'{key} not found')



NONE = 0
GET = 1
LIST = 1 << 1
CREATE = 1 << 2
UPDATE = 1 << 3
DELETE = 1 << 4
QUERIES = GET | LIST
MUTATIONS = CREATE | UPDATE | DELETE
QUERIES_AND_MUTATIONS = QUERIES | MUTATIONS
ALL = (1 << 5) | QUERIES_AND_MUTATIONS



from polecat.db.schema.variable import SessionVariable
from psycopg2.extensions import AsIs, adapt, register_adapter


def adapt_session_variable(sv):
    type = sv.type or ''
    if type:
        # TODO: Should be using a mapping from Python types to DB
        # types.
        type = f'::{type}'
    return AsIs(f"current_setting('{sv.name}', TRUE){type}")


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y


def adapt_point(point):
    # TODO: Would prefer not to have to decode.
    x = adapt(point.x).getquoted().decode()
    y = adapt(point.y).getquoted().decode()
    return AsIs("'(%s, %s)'" % (x, y))


register_adapter(Point, adapt_point)
register_adapter(SessionVariable, adapt_session_variable)



import click
from termcolor import colored

from .feedback import HaloFeedback
from .main import main

__all__ = ('deploy', 'undeploy')


@main.command()
@click.option('--deployment')
@click.option('--dry-run', is_flag=True)
@click.pass_context
def deploy(ctx, deployment, dry_run):
    from ..deploy.aws.deploy import deploy as aws_deploy
    project = ctx.obj['project']
    bucket = ctx.obj['bucket']
    urls = aws_deploy(project, bucket, deployment, dry_run, feedback=HaloFeedback())
    for deployment, url in urls.items():
        print(f'  {deployment}: {colored(url, "green")}')


@main.command()
@click.option('--deployment')
@click.pass_context
def undeploy(ctx, deployment):
    from ..deploy.aws.deploy import undeploy as aws_undeploy
    project = ctx.obj['project']
    aws_undeploy(project, deployment, feedback=HaloFeedback())



import click

from .main import main

__all__ = ('repo', 'project', 'app',)


@main.group()
@click.pass_context
def start(ctx):
    pass


@start.command()
@click.argument('name')
async def repo(name):
    pass


@start.command()
@click.argument('name')
async def project(name):
    pass


@start.command()
@click.argument('name')
async def app(name):
    pass



class Input:
    def __init__(self, type=None, input=None):
        if type is not None and input is not None:
            self.translate(type, input)

    def translate(self, type, input):
        self.change = {}
        self.delete = {}
        for cc_name, graphql_field in type.fields.items():
            field = graphql_field._field
            change, delete = field.from_input(input, type)
            if change:
                self.change.update(change)
            if delete:
                self.merge_delete(delete)

    def merge_delete(self, delete, source=None):
        if not source:
            source = self.delete
        for model, ids in delete.items():
            source.setdefault(model, set()).update(ids)


def parse_id(id):
    if id is None:
        return id
    try:
        return int(id)
    except ValueError:
        raise ValueError(f'Cannot parse ID: {id}')



from collections import defaultdict

from ..schema import RelatedColumn, ReverseColumn
from ..schema.variable import SessionVariable
from .selection import Selection


class Queryable:
    selectable = True
    mutatable = True


class Query(Queryable):
    def __init__(self, source=None):
        self.source = source

    @property
    def root(self):
        # TODO: I probably want to cache this?
        return getattr(self.source, 'root', self.source)

    def assert_selectable(self, queryable):
        if not getattr(queryable, 'selectable', False):
            raise ValueError(f'Queryable is not selectable {queryable}')

    def assert_mutatable(self, queryable):
        if not getattr(queryable, 'mutatable', False):
            raise ValueError(f'Queryable is not mutatable {queryable}')


class Select(Query):
    mutatable = False

    def __init__(self, source, selection, limit=None, order=None, **kwargs):
        super().__init__(source, **kwargs)
        self.assert_selectable(source)
        self.selection = selection
        self.limit = limit
        self.order = order
        self.recurse_column = None

    def iter_column_names(self):
        return iter(self.selection)

    def has_column(self, name):
        return self.selection.has_column(name)


class Insert(Query):
    mutatable = False

    def __init__(self, source, values, **kwargs):
        super().__init__(source, **kwargs)
        self.reverse_queries = defaultdict(list)
        values = self.coerce_values(values)
        self.assert_mutatable(source)
        self.assert_selectable(values)
        self.assert_values_match_columns(values)
        self.values = values

    def iter_column_names(self):
        return iter(self.source.iter_column_names())

    def has_column(self, name):
        return self.source.has_column(name)

    def get_column(self, name):
        return self.source.get_column(name)

    def coerce_values(self, values):
        if isinstance(values, dict):
            return self.dict_to_values(values)
        else:
            return values

    def dict_to_values(self, values):
        mapped_values = {}
        for column_name, subvalue in values.items():
            column = self.get_column(column_name)
            # TODO: It would be nice to avoid breaking the OCP here.
            if isinstance(column, RelatedColumn):
                subvalue = self.subvalues_to_subquery(column, subvalue)
            elif isinstance(column, ReverseColumn):
                self.create_reverse_queries(column, subvalue)
                continue
            mapped_values[column_name] = subvalue
        return Values(mapped_values)

    def subvalues_to_subquery(self, column, subvalues):
        if isinstance(subvalues, (Query, SessionVariable)):
            return subvalues
        # TODO: Ugh, this is soooo bad. This is how I'm currently
        # handing being given an integer for the primary key.
        if isinstance(subvalues, int):
            return subvalues
        if subvalues is None:
            return subvalues
        # TODO: Should we be considering if there's an ID in the
        # subset? If so this should be an Update instead of a Insert.
        return Select(
            Insert(column.related_table, subvalues),
            Selection('id')
        )

    def create_reverse_queries(self, column, linked_values):
        related_column = column.related_column
        for values in linked_values:
            reverse_query = Insert(
                column.related_table,
                {
                    **values,
                    related_column.name: Select(self, Selection('id'))
                }
            )
            self.reverse_queries[column.name].append(reverse_query)

    def assert_values_match_columns(self, values):
        for column_name in values.iter_column_names():
            # TODO: This assumes the source has the `has_column`
            # method. Should we formalise this in an interface?
            if not self.source.has_column(column_name):
                raise ValueError(f'Invalid column "{column_name}" name in insert')


class InsertIfMissing(Insert):
    def __init__(self, source, values, defaults, **kwargs):
        super().__init__(source, values, **kwargs)
        self.defaults = defaults


class Update(Insert):
    def assert_mutatable(self, source):
        # TODO: Think about this more, but Update can work with pretty
        # much anything. Maybe need "is_insertable" and "is_updatable"?
        return True


class Delete(Query):
    mutatable = False

    def __init__(self, source, **kwargs):
        super().__init__(source, **kwargs)
        # TODO: Need to remove this to support filters.
        # self.assert_mutatable(source)

    def iter_column_names(self):
        return iter(self.source.iter_column_names())

    def has_column(self, name):
        return self.source.has_column(name)


class Filter(Query):
    mutatable = False

    def __init__(self, source, expression=None, options=None, **kwargs):
        super().__init__(source, **kwargs)
        self.assert_selectable(source)
        self.expression = expression
        self.options = options or {}

    def has_column(self, name):
        return self.source.has_column(name)

    def get_column(self, name):
        return self.source.get_column(name)


class Join(Query):
    mutatable = False

    def __init__(self, source, separator=' ', **kwargs):
        super().__init__(source, **kwargs)
        self.separator = separator


class Common(Query):
    mutatable = False

    def __init__(self, subqueries):
        super().__init__()
        self.subqueries = subqueries


class Values(Query):
    mutatable = False

    def __init__(self, values, columns=None, **kwargs):
        super().__init__(**kwargs)
        if isinstance(values, dict):
            self.columns = tuple(values.keys())
            self.values = (tuple(values.values()),)
        else:
            self.columns = columns
            self.values = values

    def iter_rows(self):
        for values in self.values:
            yield zip(self.columns, values)

    def iter_column_names(self):
        return iter(self.columns)

    def is_bulk(self):
        return len(self.values) > 1

    def as_dict(self):
        assert len(self.values) == 1
        return dict(zip(self.columns, self.values[0]))



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from polecat.db.migration import (bootstrap_migrations, diff_schemas,
                                  make_migrations, migrate)
from polecat.db.schema import (CASCADE, Column, IntColumn, RelatedColumn,
                               Schema, Table)

from ..models import schema  # noqa


@pytest.mark.skip(reason='Conflicts between projects and migrations')
def test_migration_from_models(testdb):
    bootstrap_migrations()
    migrations = diff_schemas(schema)
    migrations = [
        *migrations
    ]
    for mgr in migrations:
        # for op in mgr.operations:
        #     sql = testdb.mogrify(*op.sql)
        #     print(sql.decode())
        mgr.forward()


def test_serialize_migration(testdb):
    migrations = diff_schemas(schema)
    for mgr in migrations:
        result = mgr.serialize()
        # print(result)
        # TODO: Better test.
        assert result is not None
        assert len(result) > 0


def test_run_migrations(testdb):
    bootstrap_migrations()
    migrations = diff_schemas(schema)
    apps = []
    with TemporaryDirectory() as root:
        for mgr in migrations:
            if mgr.app:
                mgr.app.path = Path(root) / mgr.app.name
                apps.append(mgr.app)
        for mgr in migrations:
            mgr.save(root)
        migrate([root], apps=apps)
        # TODO: Test something?


def test_make_migrations(testdb):
    bootstrap_migrations()
    with TemporaryDirectory() as root:
        make_migrations(to_schema=schema, output_path=root)


@pytest.mark.skip(reason='need to rebuild this')
def test_no_new_migrations(testdb):
    bootstrap_migrations()
    with TemporaryDirectory() as root:
        migrations = make_migrations(
            to_schema=schema,
            output_path=root,
            apps=[type('App', (), {
                'name': 'auth',
                'path': Path(root) / 'auth'
            })],
            migration_paths=[root]
        )
        assert len(migrations) != 0
        migrations = make_migrations(
            to_schema=schema,
            output_path=root,
            apps=[type('App', (), {
                'name': 'auth',
                'path': Path(root) / 'auth'
            })],
            migration_paths=[root]
        )
        assert len(migrations) == 0


@pytest.mark.skip(reason='need to mock app registry')
def test_dependencies(testdb):
    bootstrap_migrations()
    schema = Schema(tables=[
        Table('t0', columns=[
            Column('id', 'int', unique=True),
            RelatedColumn('other', 'int', 'a1.t1.id')
        ], app='a0'),
        Table('t1', columns=[
            Column('id', 'int', unique=True)
        ], app='a1')
    ])
    migrations = schema.diff()
    assert len(migrations) == 2
    for mgr in migrations:
        mgr.forward()


def test_cascade_constraint(testdb):
    t0 = Table('t0', columns=[
        IntColumn('id', 'int', primary_key=True)
    ])
    t1 = Table('t1', columns=[
        IntColumn('id', 'int', primary_key=True),
        RelatedColumn('t0', t0, 'int', on_delete=CASCADE)
    ])
    schema = Schema(tables=[t0, t1])
    schema.bind()
    migrations = diff_schemas(schema)
    sql = testdb.mogrify(*migrations[0].operations[1].sql)
    assert sql == (
        b'CREATE TABLE "t1" (\n"id" int PRIMARY KEY,\n"t0" int REFERENCES'
        b' "t0"("id") ON DELETE CASCADE \n);\n'
    )



from polecat.db.connection import transaction
from polecat.model.db import Q

from ..models import Movie


def test_transaction(migrateddb, factory):
    n_movies_pre = len(list(Q(Movie).select()))
    try:
        with transaction():
            factory.Movie()
            raise Exception
    except Exception:
        pass
    n_movies_post = len(list(Q(Movie).select()))
    assert n_movies_pre == n_movies_post



# TODO: Migration shouldn't really be aware of SQL specifics, they
# should live in "polecat.db.sql".
import polecat.db.sql.postgres  # noqa
from psycopg2.sql import SQL, Identifier

from ...core.context import active_context
from ...utils import indent
from ...utils.predicates import not_empty
from ..decorators import dbcursor
from ..schema import Access, MutableColumn, Role, Table
from .column import Column  # noqa

create_role_sql = '''do $$
  BEGIN
    IF NOT EXISTS(SELECT FROM pg_catalog.pg_roles WHERE rolname = %s) THEN
      CREATE ROLE {role};
    END IF;
    GRANT {role} TO CURRENT_USER;
    GRANT USAGE, SELECT ON ALL SEQUENCES IN SCHEMA public TO {role};
    ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT USAGE, SELECT ON SEQUENCES TO {role};
  END;
$$;'''

create_table_template = '''operation.CreateTable(
    '{}',
    columns=[{}]{}
)'''


class Operation:
    def __init__(self):
        self.post_ops = []

    @property
    def dependencies(self):
        return ()

    @dbcursor
    def forward(self, schema, cursor):
        cursor.execute(*self.sql)

    def forward_schema(self, schema):
        pass

    def set_app(self, app):
        pass


class CreateExtension(Operation):
    def __init__(self, name):
        super().__init__()
        self.name = name

    @property
    def signature(self):
        return (CreateExtension, self.name)

    @property
    def sql(self):
        return (
            SQL('CREATE EXTENSION IF NOT EXISTS {};').format(
                Identifier(self.name)
            ),
            ()
        )

    def serialize(self):
        return f"CreateExtension('{self.name}')"


class CreateTable(Operation):
    def __init__(self, name_or_table, *args, **kwargs):
        super().__init__()
        if not isinstance(name_or_table, Table):
            # TODO: This is super sketchy. What if the app name changes?
            name_or_table = name_or_table.split('_')[-1]
            name_or_table = Table(name_or_table, *args, **kwargs)
        self.table = name_or_table

    @property
    def signature(self):
        return self.table.signature

    @property
    def app(self):
        return self.table.app

    @property
    def dependencies(self):
        for col in self.table.columns:
            for dep in col.dependencies:
                yield dep

    @property
    def sql(self):
        table_name_ident = Identifier(self.get_table_name())
        all_sql, all_args = self.all_sql()
        self.post_ops.append(self.indexes_sql())
        self.post_ops.append(self.policies_sql())
        return (
            SQL('\n').join(
                (SQL('CREATE TABLE {} (\n{}\n);').format(
                    table_name_ident,
                    all_sql
                ),) +
                self.access_sql(table_name_ident)
            ),
            all_args
        )

    def set_app(self, app):
        self.table.app = app

    def forward_schema(self, schema):
        schema.add_table(self.table)
        schema.bind_table(self.table)

    def all_sql(self):
        columns_sql, columns_args = self.all_columns_sql()
        return SQL(',\n').join(tuple(filter(not_empty, (
            columns_sql,
            self.all_uniques_sql(),
            self.all_checks_sql()
        )))), columns_args

    def all_columns_sql(self):
        all_sql = []
        all_args = ()
        for col in self.iter_valid_columns():
            sql, args = self.column_sql(col)
            all_sql.append(sql)
            all_args += args
        return SQL(',\n').join(all_sql), all_args

    def all_uniques_sql(self):
        uniques = self.table.uniques
        if len(uniques):
            return SQL(',\n').join(
                SQL('  UNIQUE ({})').format(self.unique_sql(u))
                for u in uniques
            )
        else:
            None

    def all_checks_sql(self):
        checks = self.table.checks
        if len(checks):
            return SQL(',\n').join(
                SQL('  CHECK ({})').format(self.check_sql(c))
                for c in checks
            )
        else:
            None

    @active_context
    def iter_valid_columns(self, context):
        registry = context.registries.migration_column_registry
        for column in self.table.columns:
            if isinstance(column, MutableColumn):
                yield registry[column](column)

    def column_sql(self, column):
        return column.to_sql()

    def check_sql(self, check):
        if isinstance(check, str):
            return SQL(check)
        raise NotImplemented('checks must be strings')

    def unique_sql(self, unique):
        if isinstance(unique, (tuple, list)):
            return SQL(',').join(map(Identifier, unique))
        raise NotImplemented('uniques must be iterable')

    def access_sql(self, table_name):
        access = self.table.access
        if access:
            grants = tuple(self.iter_grants(access, table_name))
        else:
            grants = ()
        return grants

    def indexes_sql(self):
        all_sql, all_args = [], ()
        for index in self.table.indexes:
            sql, args = index.sql
            all_sql.append(sql)
            all_args += args
        return SQL('\n').join(all_sql), all_args

    def policies_sql(self):
        all_sql, all_args = [], ()
        for policy in self.table.policies:
            sql, args = policy.sql
            all_sql.append(sql)
            all_args += args
        return SQL('\n').join(all_sql), all_args

    def iter_grants(self, access, table_name):
        roles_and_perms = (
            (access.all, 'ALL'),
            (access.select, 'SELECT'),
            (access.insert, 'INSERT'),
            (access.update, 'UPDATE'),
            (access.delete, 'DELETE')
        )
        for roles, perm in roles_and_perms:
            if roles:
                for role in roles:
                    yield SQL('GRANT %s ON {} TO {};' % perm).format(
                        table_name,
                        Identifier(role.Meta.dbrole.dbname)  # TODO: Should already be DB roles.
                    )

    def get_table_name(self):
        # TODO: Same operation in models/registry.py.
        # TODO: Also, not sure this should even exist... should I be
        # just setting the actual table name instead of the model
        # name?
        return self.table.name

    def serialize(self):
        columns = self.serialize_valid_columns()
        if columns:
            columns = indent(f'\n{columns}', 8)
            columns += indent('\n')
        options = []
        access = self.serialize_access()
        if access:
            options.append(f'access={access}')
        checks = self.serialize_checks()
        if checks:
            options.append(f'checks={checks}')
        uniques = self.serialize_uniques()
        if uniques:
            options.append(f'uniques={uniques}')
        indexes = self.serialize_indexes()
        if indexes:
            options.append(f'indexes={indexes}')
        policies = self.serialize_policies()
        if policies:
            options.append(f'policies={policies}')
        options = ',\n'.join(options)
        if options:
            options = indent(f',\n{options}', 4)
        return create_table_template.format(
            self.table.name,
            columns,
            options
        )

    def serialize_valid_columns(self):
        cols = []
        for col in self.iter_valid_columns():
            cols.append(self.serialize_column(col.schema_column))
        return ',\n'.join(cols)

    def serialize_column(self, column):
        cargs = column.get_construction_arguments()
        return f'column.{column.__class__.__name__}({str(cargs)})'

    def serialize_access(self):
        access = self.table.access
        if access:
            access = {
                'all': [a.Meta.role for a in access.all],
                'select': [a.Meta.role for a in access.select],
                'insert': [a.Meta.role for a in access.insert],
                'update': [a.Meta.role for a in access.update],
                'delete': [a.Meta.role for a in access.delete]
            }
            access = {
                k: v
                for k, v in access.items()
                if len(v) > 0
            }
            # TODO: Maybe try make it indented and pretty?
            access = str(access)
        return access

    def serialize_checks(self):
        checks = self.table.checks
        if checks:
            checks = str(checks)
        return checks

    def serialize_uniques(self):
        uniques = self.table.uniques
        if uniques:
            uniques = str(uniques)
        return uniques

    def serialize_indexes(self):
        indexes = self.table.indexes
        if indexes:
            indexes = '[' + ', '.join(
                index.serialize()
                for index in indexes
            ) + ']'
        return indexes

    def serialize_policies(self):
        policies = self.table.policies
        if policies:
            policies = '[' + ', '.join(
                policy.serialize()
                for policy in policies
            ) + ']'
        return policies


class DeleteTable(Operation):
    pass


class AlterTable(Operation):
    pass


class CreateRole(Operation):
    def __init__(self, name_or_role, *args, **kwargs):
        super().__init__()
        if not isinstance(name_or_role, Role):
            name_or_role = Role(name_or_role, *args, **kwargs)
        self.role = name_or_role

    @property
    def dependencies(self):
        return self.role.parents

    @property
    def signature(self):
        return self.role.signature

    @property
    def app(self):
        return self.role.app

    @property
    def sql(self):
        role_name = self.get_role_name()
        role_name_ident = Identifier(role_name)
        return (
            SQL('\n').join(
                (SQL(create_role_sql).format(role=role_name_ident),) +
                self.grants_sql(role_name_ident)
            ),
            (role_name,)
        )

    def set_app(self, app):
        self.role.app = app

    def forward_schema(self, schema):
        schema.add_role(self.role)
        schema.bind_role(self.role)

    def grants_sql(self, role_name):
        return tuple(
            SQL('GRANT {} TO {};').format(
                role_name,
                Identifier(getattr(parent, 'dbname', parent))
            )
            for parent in self.role.parents
        )

    def get_role_name(self):
        return self.role.dbname

    def serialize(self):
        parents = self.serialize_role_parents(self.role)
        if parents:
            parents = f', parents={parents}'
        return f"operation.CreateRole('{self.role.name}'{parents})"

    def serialize_role_parents(self, role):
        parents = '[' + ', '.join(f"'{p.name}'" for p in role.parents) + ']'
        return parents


class DeleteRole(Operation):
    pass


class AlterRole(Operation):
    pass


class GrantAccess(Operation):
    def __init__(self, entity_or_access, *args, **kwargs):
        super().__init__()
        if not isinstance(entity_or_access, Access):
            entity_or_access = Access(entity_or_access, *args, **kwargs)
        self.access = entity_or_access

    @property
    def signature(self):
        return (GrantAccess, self.access.signature)

    @property
    def app(self):
        return self.access.app

    @property
    def sql(self):
        # TODO: Awful.
        access_sql = filter(not_empty, [
            self.get_access_sql(access, roles)
            for access, roles in [
                ('SELECT', set(map(lambda r: getattr(r, 'dbname', r), self.access.all)) | set(map(lambda r: getattr(r, 'dbname', r), self.access.select))),
                ('INSERT', set(map(lambda r: getattr(r, 'dbname', r), self.access.all)) | set(map(lambda r: getattr(r, 'dbname', r), self.access.insert))),
                ('UPDATE', set(map(lambda r: getattr(r, 'dbname', r), self.access.all)) | set(map(lambda r: getattr(r, 'dbname', r), self.access.update))),
                ('DELETE', set(map(lambda r: getattr(r, 'dbname', r), self.access.all)) | set(map(lambda r: getattr(r, 'dbname', r), self.access.delete))),
            ]
        ])
        return SQL('\n').join(access_sql), ()

    @property
    def entity_sql(self):
        tag = self.get_entity_tag(self.access.entity)
        return SQL('%s {}' % tag).format(
            Identifier(self.access.entity if isinstance(self.access.entity, str) else self.access.entity.name)
        )

    def set_app(self, app):
        self.access.app = app

    def forward_schema(self, schema):
        raise NotImplementedError

    def get_entity_tag(self, entity):
        raise NotImplementedError

    def get_access_sql(self, access, roles):
        if not roles:
            return None
        return SQL('GRANT %s ON {} TO {};' % access).format(
            self.entity_sql,
            SQL(', ').join([
                Identifier(r if isinstance(r, str) else r.dbname)  # TODO: Explain
                for r in roles
            ])
        )

    def serialize(self):
        # TODO: Fix this.
        entity = self.serialize_entity()
        access = [
            ('all', [r.name for r in self.access.all]),
            ('select', [r.name for r in self.access.select]),
            ('insert', [r.name for r in self.access.insert]),
            ('update', [r.name for r in self.access.update]),
            ('delete', [r.name for r in self.access.delete])
        ]
        access = ', '.join([f'{a[0]}={a[1]}' for a in access if len(a[1])])
        return f'operation.{self.__class__.__name__}({repr(entity)}, {access})'

    def serialize_entity(self):
        entity = self.access.entity
        if isinstance(entity, str):
            entity_string = entity
        else:
            entity_string = self.entity.name
        return entity_string


class GrantAccessToTable(GrantAccess):
    def serialize_entity(self):
        entity = self.access.entity
        if isinstance(entity, str):
            entity_string = entity
        else:
            entity_string = entity.name
        return entity_string

    def get_entity_tag(self, entity):
        return 'TABLE'

    def forward_schema(self, schema):
        if isinstance(self.access.entity, str):
            self.access.entity = schema.get_table_by_name(self.access.entity)
        schema.add_access(self.access)
        schema.bind_access(self.access)


class GrantAccessToRole(GrantAccess):
    def get_entity_tag(self, entity):
        return 'ROLE'

    def forward_schema(self, schema):
        self.access.entity = schema.get_role_by_name(self.access.entity)
        schema.add_access(self.access)
        schema.bind_access(self.access)


class RevokeAccess(Operation):
    def __init__(self, entity_or_access, *args, **kwargs):
        super().__init__()
        if not isinstance(entity_or_access, Access):
            entity_or_access = Access(entity_or_access, *args, **kwargs)
        self.access = entity_or_access

    @property
    def signature(self):
        return (RevokeAccess, self.access.signature)

    @property
    def app(self):
        return self.access.app

    def set_app(self, app):
        self.access.app = app

    def forward_schema(self, schema):
        schema.remove_access(self.access)


class RunPython(Operation):
    def __init__(self, forward_func):
        super().__init__()
        self.forward_func = forward_func

    @dbcursor
    def forward(self, schema, cursor):
        self.forward_func(schema, cursor)



import click
from termcolor import colored

from .feedback import HaloFeedback
from .main import main

__all__ = ('create_deployment', 'list_deployments')


@main.group()
@click.pass_context
def deployment(ctx):
    pass


@deployment.command('create')
@click.argument('deployment')
@click.pass_context
def create_deployment(ctx, deployment):
    from ..deploy.aws.deployment import create_deployment as aws_create_deployment
    project = ctx.obj['project']
    aws_create_deployment(project, deployment, feedback=HaloFeedback())


@deployment.command('list')
@click.pass_context
def list_deployments(ctx):
    from ..deploy.aws.deployment import list_deployments as aws_list_deployments
    project = ctx.obj['project']
    deployments = aws_list_deployments(project, feedback=HaloFeedback())
    if len(deployments) == 0:
        print(colored('    No deployments', 'grey'))
    else:
        for dep in deployments:
            print(f'    {dep}')


@deployment.command('delete')
@click.argument('deployment')
@click.pass_context
def delete_deployment(ctx, deployment):
    from ..deploy.aws.deployment import delete_deployment as aws_delete_deployment
    project = ctx.obj['project']
    aws_delete_deployment(project, deployment, feedback=HaloFeedback())



from . import to_tuple
from .optiondict import Option, OptionDict
from .passthrough import passthrough


class MetaMetaclass(type):
    def __new__(metaclass, name, bases, attrs):
        if name == 'Meta':
            return super().__new__(metaclass, name, bases, attrs)
        return metaclass.build_dict_type(metaclass, name, attrs)

    def build_dict_type(metaclass, name, attrs, base=OptionDict):
        cls = super().__new__(metaclass, name, (base,), attrs)
        cls.initial_options = build_options(attrs)
        return passthrough(cls)


def build_options(attrs):
    options = {}
    for key, option in attrs.items():
        if key.startswith('__'):
            continue
        if not isinstance(option, Option):
            args = (key,) + to_tuple(option)
            option = Option.factory(*args)
        else:
            option.key = key
        options[option.key] = option
    return options


class Meta(metaclass=MetaMetaclass):
    class Any:
        pass



from .command import Command


class Migrate(Command):
    def get_params(self):
        return (
            self.Option(('--skip',), is_flag=True, default=False),
        )

    def run(self, skip):
        from polecat.db.migration import migrate
        migrate(
            apps=self.project.apps,
            migration_paths=(self.project.path,)
        )



from polecat.db.schema import CASCADE
from polecat.db.schema.utils import Auto
from psycopg2.sql import SQL, Identifier

from ...core.registry import MappedRegistry, RegistryMetaclass
from .. import schema

__all__ = ('Column', 'IntColumn')

MappedRegistry('migration_column_registry')


class Column(metaclass=RegistryMetaclass):
    _registry = 'migration_column_registry'
    dbtype = None

    def __init__(self, schema_column):
        self.schema_column = schema_column

    def to_sql(self):
        col = self.schema_column
        constraints_sql, args = self.get_constraints_sql()
        parts = [
            Identifier(col.name),
            self.get_type_with_dimensions_sql(),
            constraints_sql
        ]
        return SQL(' ').join(parts), args

    def get_constraints_sql(self):
        constraints = []
        args = ()
        col = self.schema_column
        if col.primary_key:
            constraints.append('PRIMARY KEY')
        else:
            if not col.null:
                constraints.append('NOT NULL')
            if col.unique:
                constraints.append('UNIQUE')
            if col.default is not None:
                def_str, def_args = self.get_default_str()
                constraints.append(def_str)
                args += def_args
        return SQL(' '.join(constraints)), args

    def get_default_str(self):
        col = self.schema_column
        return (
            'DEFAULT %s',
            (self.translate_default_value(col.default),)
        )

    def get_type_with_dimensions_sql(self):
        col = self.schema_column
        sql = self.get_type_sql()
        if col.dimensions:
            sql = SQL('{}%s' % '[]'*col.dimensions).format(
                sql
            )
        return sql

    def get_type_sql(self):
        return SQL(self.dbtype)

    def translate_default_value(self, value):
        return value


class IntColumn(Column):
    sources = (schema.IntColumn,)
    dbtype = 'int'


class FloatColumn(Column):
    sources = (schema.FloatColumn,)
    dbtype = 'real'


class PointColumn(Column):
    sources = (schema.PointColumn,)
    dbtype = 'point'


class SerialColumn(Column):
    sources = (schema.SerialColumn,)
    dbtype = 'serial'


class TextColumn(Column):
    sources = (schema.TextColumn,)
    dbtype = 'text'

    def get_type_sql(self):
        col = self.schema_column
        if col.max_length:
            return SQL(f'varchar({col.max_length})')
        else:
            return SQL(self.dbtype)


class PasswordColumn(TextColumn):
    sources = (schema.PasswordColumn,)
    dbtype = 'chkpass'


class DateColumn(Column):
    sources = (schema.DateColumn,)
    dbtype = 'date'

    def get_default_str(self):
        col = self.schema_column
        if col.default == Auto:
            return 'DEFAULT now()', ()
        else:
            return super().get_default_str()


class TimestampColumn(Column):
    sources = (schema.TimestampColumn,)
    dbtype = 'timestamptz'

    def get_default_str(self):
        col = self.schema_column
        if col.default == Auto:
            return 'DEFAULT now()', ()
        else:
            return super().get_default_str()


class UUIDColumn(Column):
    sources = (schema.UUIDColumn,)
    dbtype = 'uuid'

    def get_default_str(self):
        col = self.schema_column
        if col.default == Auto:
            return 'DEFAULT uuid_generate_v4()', ()
        else:
            return super().get_default_str()


class BoolColumn(Column):
    sources = (schema.BoolColumn,)
    dbtype = 'boolean'


class JSONColumn(Column):
    sources = (schema.JSONColumn,)
    dbtype = 'jsonb'


class RelatedColumn(IntColumn):
    sources = (schema.RelatedColumn,)

    def to_sql(self):
        col = self.schema_column
        constraints_sql, args = self.get_constraints_sql()
        parts = [
            Identifier(col.name),
            self.get_type_sql(),
            self.get_reference_sql(),
            constraints_sql
        ]
        return SQL(' ').join(parts), args

    def get_reference_sql(self):
        col = self.schema_column
        return SQL('REFERENCES {}({}){}').format(
            Identifier(getattr(col.related_table, 'name', col.related_table)),
            Identifier('id'),
            self.get_on_delete_sql()
        )

    def get_on_delete_sql(self):
        col = self.schema_column
        if col.on_delete == CASCADE:
            sql = SQL(' ON DELETE CASCADE')
        else:
            sql = SQL('')
        return sql



from datetime import timedelta

import pytimeparse
from polecat.utils import Undefined, to_bool, to_dict, to_list


class UndefinedOptionError(Exception):
    pass


class DuplicateOptionError(Exception):
    pass


class OptionDict:
    """ A dictionary-like class in which all possible options are
    specified in advance. Allows for acknowledgement of incorrect
    accesses and suggestions for the correct key.
    """
    def __init__(self, options=None):
        self.options = {**getattr(self, 'initial_options', {})}
        self.values = {}
        self.add_options(*(options or ()))

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def add_option(self, option):
        return self.add_options(option)

    def add_options(self, *options):
        for opt in options:
            if opt.key in self.options:
                raise DuplicateOptionError(f'{opt.key} already exists')
            self.options[opt.key] = opt
        return self

    def get(self, key):
        try:
            return self.get_option(key).get(self)
        except UndefinedOptionError:
            return self.handle_undefined(key)

    def set(self, key, value):
        self.get_option(key).set(self, value)

    def has_option(self, key):
        return key in self.options

    def get_option(self, key):
        try:
            return self.options[key]
        except KeyError:
            raise self.key_error(key)

    def handle_undefined(self, key):
        raise

    def merge(self, other):
        for key, value in other.items():
            self.get_option(key).merge(self, value)

    def merge_from_type(self, other):
        for option in self.options.values():
            option.merge_from_type(self, other)

    def items(self):
        for key, option in self.options.values():
            yield key, option.get()

    def remove_options(self, *options):
        for opt in options:
            try:
                del self.options[opt]
            except KeyError:
                pass
        return self

    def get_default(self, option):
        raise UndefinedOptionError(f'{option.key} is undefined')

    def key_error(self, key):
        try:
            from Levenshtein import distance
        except ImportError:
            return KeyError(f'{key} not found')
        closest = [
            (distance(key, opt), opt)
            for opt in self.options
        ]
        if closest:
            closest = min(closest, key=lambda x: x[0])[1]
            return KeyError(f'{key} not found, did you mean {closest}?')
        else:
            return KeyError(f'{key} not found')


class OptionMetaclass(type):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        type = getattr(cls, 'TYPE', None)
        if type is not None:
            cls.TYPE_MAP[type] = cls


class Option(metaclass=OptionMetaclass):
    TYPE_MAP = {}

    @staticmethod
    def factory(option_or_key, type=None, default=Undefined):
        if isinstance(option_or_key, Option):
            return option_or_key
        option_class = Option.TYPE_MAP.get(type, Option)
        return option_class(option_or_key, default=default)

    def __init__(self, key=None, default=Undefined):
        self.key = key
        self.default = default

    def get(self, obj):
        value = self.get_value(obj)
        if value is Undefined:
            if self.default is Undefined:
                return obj.get_default(self)
            return self.default
        return value

    def set(self, obj, value):
        self.set_value(obj, self.coerce(value))

    def get_value(self, obj):
        return obj.values.get(self.key, Undefined)

    def set_value(self, obj, value):
        obj.values[self.key] = value

    def coerce(self, value):
        return value

    def merge(self, obj, other):
        self.set(obj, other)

    def merge_from_type(self, obj, other):
        value = getattr(other, self.key, Undefined)
        if value is not Undefined:
            self.merge(obj, value)


class IntOption(Option):
    TYPE = int

    def coerce(self, value):
        return int(value)


class StrOption(Option):
    TYPE = str

    def coerce(self, value):
        return str(value)


class BoolOption(Option):
    TYPE = bool

    def coerce(self, value):
        return to_bool(value)


class TimedeltaOption(Option):
    TYPE = timedelta

    def coerce(self, value):
        if isinstance(value, str):
            value = timedelta(pytimeparse.parse(value))
        return value


class ListOption(Option):
    TYPE = list

    def coerce(self, value):
        return to_list(value)

    def merge(self, obj, other):
        return self.get_value(obj).extend(self.coerce(other))


class DictOption(Option):
    TYPE = dict

    def coerce(self, value):
        return to_dict(value)

    def merge(self, obj, other):
        return self.get_value(obj).merge(self.coerce(other))



from ..query import query as query_module
from .expression.insert import Insert
from .expression.subquery import Subquery
from .expression.subrelation_override import SubrelationOverride
from .expression.values import Values


class InsertStrategy:
    def __init__(self, root_strategy):
        self.root = root_strategy
        self.relation_counter = 0

    def parse_query(self, query):
        expr = Insert(
            query.source,
            self.parse_values_or_subquery(query, query.values)
        )
        if query.reverse_queries:
            expr = SubrelationOverride(expr, query.reverse_queries, self.root)
        return expr

    def parse_values_or_subquery(self, query, values_or_subquery):
        if isinstance(values_or_subquery, query_module.Values):
            return self.parse_values(query, values_or_subquery)
        else:
            return self.parse_subquery(values_or_subquery)

    def parse_values(self, query, values):
        parsed_values = []
        for row in values.iter_rows():
            row_values = []
            for _, value in row:
                if isinstance(value, query_module.Query):
                    value = self.parse_subquery(value)
                row_values.append(value)
            parsed_values.append(row_values)
        # TODO: Forced override.
        values.values = parsed_values
        return Values(values, query.source)

    def parse_subquery(self, subquery):
        if self.should_add_select(subquery):
            subquery = query_module.Select(
                subquery,
                query_module.Selection('id')
            )
        return Subquery(self.root.parse_queryable_or_builder(subquery))

    def should_add_select(self, subquery):
        return isinstance(subquery, (query_module.Insert, query_module.Update))



import os
import re
from importlib import import_module

from graphql_server import HttpQueryError
from polecat.admin.commands import *  # noqa
from polecat.core.config import default_config
from polecat.core.context import active_context
from polecat.db.role_prefix import set_role_prefix
from polecat.model import default_blueprint
from polecat.model.db.helpers import create_schema
from polecat.utils import get_class_path

from .index import IndexHandler, get_index_html

active_project = None


def get_active_project():
    global active_project
    return active_project


def set_active_project(project):
    global active_project
    active_project = project


def load_project():
    global active_project
    if active_project:
        return active_project
    module_path = os.environ.get('POLECAT_PROJECT_MODULE')
    if module_path is None:
        raise Exception('Please set POLECAT_PROJECT_MODULE')
    parts = module_path.split('.')
    module_name, class_name = '.'.join(parts[:-1]), parts[-1]
    project_class = getattr(import_module(module_name), class_name)
    project = project_class()
    active_project = project
    # proxy_project.set_target(project)
    return project


def setup_project():
    project = load_project()
    project.prepare()
    return project


class Project:
    name = None
    deployment = None
    bundle = None
    default_role = None
    config = None
    enable_cors = False

    def __init__(self, name=None, deployment=None, default_role=None, enable_cors=False, config=None):
        self.models = {}  # TODO: Will remove later.
        self.name = name or self.name
        if not self.name:
            self.name = self.__class__.__name__.lower()
            if self.name.endswith('project'):
                self.name = self.name[:-7]
        self.deployment = deployment or self.deployment
        if not self.deployment:
            self.deployment = os.environ.get('DEPLOYMENT', '')
        self.default_role = default_role or self.default_role
        self.enable_cors = enable_cors or self.enable_cors
        self.bucket = os.environ.get('BUCKET')
        self.bundle = os.environ.get('BUNDLE', self.bundle)
        self.bundle_version = os.environ.get('BUNDLE_VERSION')
        self.handlers = []
        self.config = config or self.config or {}

    @property
    def apps(self):
        # TODO: Cache.
        return list(active_context().registries.app_registry)

    @property
    def path(self):
        return get_class_path(self.__class__)

    def prepare(self):
        if getattr(self, '_prepared', False):
            return
        self._prepared = True
        set_role_prefix(f'{self.name}_{self.deployment}')
        # TODO: This should be a method on the config, but it would
        # need to have an underscore prefix, which I'm starting to
        # hate.
        for key, value in self.config.items():
            default_config[key] = value
        self.prepare_apps()
        self.prepare_blueprint()
        self.prepare_schema()
        # TODO: If not in DEBUG mode, validate settings.
        if not len(self.handlers):
            # TODO: Hmm, don't like the internal import so
            # much... better way to set default handlers? Also,
            # shouldn't be importing AWS middleware here, it should be
            # setup when selecting AWS as deployment system.
            from polecat.deploy.aws.middleware import DevelopMiddleware
            from polecat.auth.middleware import RoleMiddleware
            from polecat.graphql.api import GraphqlAPI
            self.handlers.append(
                GraphqlAPI(
                    self,
                    middleware=[
                        DevelopMiddleware(),
                        RoleMiddleware(self.default_role)
                    ]
                )
            )
        self.handlers.append(IndexHandler(self))
        from polecat.admin.handler import AdminHandler  # TODO: Ugh.
        self.handlers.append(AdminHandler(self))
        for handler in self.handlers:
            handler.prepare()
        self.index_html = self.get_index_html()
        self.prepare_resolvers()

    @active_context
    def prepare_apps(self, context=None):
        # TODO: This desperately needs to be cleaned up.
        items = []
        for app in context.registries.app_registry:
            items.append(
                r'\.'.join(app.__module__.split('.')[:-1])
            )
        if not items:
            return
        prog = re.compile('(' + ')|('.join(items) + ')')
        for model in default_blueprint.iter_models():
            module_name = r'.'.join(model.__module__.split('.')[:-1])
            match = prog.search(module_name)
            try:
                app = context.registries.app_registry[match.lastindex - 1]
            except AttributeError:
                # TODO: Should we require apps?
                # raise Exception(f'model {model.Meta.name} has no app')
                # TODO: I should be able to remove this once I get
                # models into the app context.
                self.models[model.Meta.name] = model
                continue
            app.models.append(model)
            model.Meta.app = app
            # TODO: I should be able to remove this once I get models
            # into the app context.
            self.models[f'{app.name}.{model.Meta.name}'] = model
        for type in default_blueprint.iter_types():
            module_name = r'.'.join(type.__module__.split('.')[:-1])
            match = prog.search(module_name)
            try:
                app = context.registries.app_registry[match.lastindex - 1]
            except AttributeError:
                # TODO: Should we require apps?
                continue
            app.types.append(model)
            type.Meta.app = app
        for role in default_blueprint.iter_roles():
            module_name = r'.'.join(role.__module__.split('.')[:-1])
            match = prog.search(module_name)
            try:
                app = context.registries.app_registry[match.lastindex - 1]
            except AttributeError:
                # TODO: Should we require apps?
                continue
            app.roles.append(model)
            role.Meta.app = app
        for access in default_blueprint.iter_access():
            module_name = r'.'.join(access.__module__.split('.')[:-1])
            match = prog.search(module_name)
            try:
                app = context.registries.app_registry[match.lastindex - 1]
            except AttributeError:
                # TODO: Should we require apps?
                continue
            app.access.append(access)
            access.app = app

    def prepare_blueprint(self):
        # TODO: Should this go before prepare apps? Would love to
        # automatically get the app into the models etc.
        default_blueprint.run_hooks()

    @active_context
    def prepare_schema(self, context=None):
        context.db.schema = create_schema()

    def get_index_html(self):
        kwargs = {}
        if self.bundle is not None:
            kwargs['bundle'] = self.bundle
        return get_index_html(
            bucket=self.bucket,
            project=self.name,
            bundle_version=self.bundle_version,
            **kwargs
        )

    def prepare_resolvers(self):
        pass

    async def handle_event(self, event):
        for handler in self.handlers:
            if not handler.match(event):
                continue
            return await handler.run(event)
        return await self.default_handler(event)

    async def default_handler(self, event):
        if event.is_http():
            # TODO: Should use a middle-man exception so JSON-API
            # doesn't need to import GraphQL.
            raise HttpQueryError(404, 'Page not found')
        else:
            raise Exception('unhandled event')



import click

from ..deploy.server.server import Server
from ..graphql import build_graphql_schema
from ..graphql.utils import print_schema
from .main import main

__all__ = ('schema',)


@main.group()
@click.pass_context
def graphql(ctx):
    pass


@graphql.command()
def schema():
    Server()
    schema = build_graphql_schema()
    print_schema(schema)



import inspect
import copy
from contextlib import contextmanager

from polecat.core.config import default_config
from polecat.db.connection import cursor, manager
from polecat.db.query.selection import Selection
from polecat.db.session import Session
from polecat.db.utils import unparse_url
from polecat.model import default_blueprint
from polecat.model.db.migrate import sync
from polecat.model.resolver import APIContext
from polecat.project.project import Project, load_project, setup_project
from polecat.test.factory import create_model_factory
from polecat.utils import random_ident

from .utils import create_database

__all__ = ('server', 'immutabledb', 'testdb', 'migrateddb', 'db', 'factory')


class ServerFixture:
    def __init__(self):
        try:
            self.project = load_project()
            # TODO: Better exception.
        except Exception:
            self.project = Project()
        self.project.prepare()

    def all_query(self, model_class, **kwargs):
        model_class = self.get_model_class(model_class)
        api_context = self.build_api_context(model_class, **kwargs)
        return model_class.Meta.all_resolver_manager(api_context)

    def get_query(self, model_class, id, **kwargs):
        model_class = self.get_model_class(model_class)
        api_context = self.build_api_context(
            model_class,
            arguments={
                'id': id
            },
            **kwargs
        )
        return model_class.Meta.all_resolver_manager(api_context)

    def create_mutation(self, model_class, input, **kwargs):
        model_class = self.get_model_class(model_class)
        api_context = self.build_api_context(
            model_class,
            input=input,
            **kwargs
        )
        return model_class.Meta.create_resolver_manager(api_context)

    def update_mutation(self, model_class, id, input, **kwargs):
        model_class = self.get_model_class(model_class)
        api_context = self.build_api_context(
            model_class,
            arguments={
                'id': id
            },
            input=input,
            **kwargs
        )
        return model_class.Meta.create_resolver_manager(api_context)

    def delete_mutation(self, model_class, id, **kwargs):
        model_class = self.get_model_class(model_class)
        api_context = self.build_api_context(
            model_class,
            input={
                'id': id
            },
            **kwargs
        )
        return model_class.Meta.delete_resolver_manager(api_context)

    def mutation(self, name, input=None, **kwargs):
        api_context = self.build_api_context(
            input=input,
            **kwargs
        )
        mutation = default_blueprint.get_mutation(name)
        if inspect.isclass(mutation):
            mutation = mutation()
        return mutation.resolve(api_context)

    def build_api_context(self, model_class=None, arguments=None, input=None,
                          selector=None, role=None, session_variables=None):
        return TestAPIContext(
            model_class,
            arguments=arguments,
            input=input,
            selector=selector,
            role=role or self.project.default_role,
            session_variables=session_variables
        )

    def get_model_class(self, model_class):
        if isinstance(model_class, str):
            return self.project.models[model_class]
        else:
            return model_class


class TestAPIContext(APIContext):
    def __init__(self, model_class=None, arguments=None, input=None,
                 selector=None, role=None, session_variables=None):
        super().__init__()
        self._model_class = model_class
        self._arguments = arguments or {}
        self._input = input or {}
        self._selector = selector or Selection()
        self.session = Session(role=role, variables=session_variables)
        self.event = type('TestEvent', (), {
            'role': role,
            'session': self.session
        })

    @property
    def model_class(self):
        return self._model_class

    def parse_argument(self, name):
        return self._arguments.get(name)

    def parse_input(self):
        return self._input

    def raw_input(self):
        return self._input

    def get_selector(self):
        return self._selector


@contextmanager
def server():
    yield ServerFixture()


@contextmanager
def immutabledb():
    with cursor() as curs:
        yield curs


@contextmanager
def testdb():
    with create_database() as curs:
        yield curs


@contextmanager
def migrateddb():
    # Should be scoped to the session.
    try:
        # Loading the project ensures we have all models.
        setup_project()
        # TODO: Catch better error.
    except Exception:
        pass
    with create_database() as curs:
        sync(cursor=curs)
        yield curs


@contextmanager
def db(migrateddb):
    # TODO: DSN representation in the psycopg2 connection has changed.
    # url = migrateddb.connection.dsn
    url = manager.get_url()
    dbinfo = manager.parse_url(url)
    local_dbname = random_ident()
    migrateddb.execute(
        f'create database {local_dbname}'
        f' with template {dbinfo["dbname"]}'
    )
    try:
        local_url = unparse_url({**dbinfo, 'dbname': local_dbname})
        with manager.push_url(local_url):
            with cursor() as curs:
                yield curs
    finally:
        with cursor() as curs:
            curs.execute(f'drop database {local_dbname}')


@contextmanager
def factory():
    # Should be scoped to the session.
    yield create_model_factory()


@contextmanager
def config(**kwargs):
    old_cfg = default_config.get_target()
    try:
        cfg = copy.deepcopy(old_cfg)
        default_config.set_target(cfg)
        for key, value in kwargs.items():
            cur_cfg = cfg
            parts = key.split('__')
            for p in parts[:-1]:
                cur_cfg = getattr(cur_cfg, p)
            setattr(cur_cfg, parts[-1], value)
        yield cfg
    finally:
        default_config.set_target(old_cfg)
